""" A kernel transformation that collects data until the vector size is reached """
""" A kernel transformation that collects data until the vector size is reached """
import pymbolic.primitives as prim
import loopy as lp
import numpy as np

from pymbolic.mapper.dependency import DependencyMapper

def collect_vector_data(knl, insns, inames, vec_size=4):
    # TODO: In theory this vec_size should be deduced from the types
    # that we find in the actual instructions

    # Process inames input -> frozenset
    if isinstance(inames, str):
        inames = inames.split(",")
    inames = frozenset(inames)

    # Filter instructions into the groups that we need for this transformation:
    # 1. Those that should be vectorized by this transformation -> insns
    # 2. Writers of quantities that depend on the inames and that 1) need -> write_insns
    # 3. "Outside" instructions, that the instructions from 1)+2) depend on -> dep_insns
    # 4. All other instructions (needed to correctly reconstruct) -> other_insns
    #
    # All these should be lists of instruction instances.

    # 1. Those that should be vectorized by this transformation -> insns
    #
    # Instructions may be passed in as either:
    # * A match expression object
    # * A comma separated string of IDs
    # * An iterable of IDs
    if isinstance(insns, lp.match.MatchExpressionBase):
        insns = lp.find_instructions(knl, insns)
    else:
        if isinstance(insns, str):
            insns = [i.strip() for i in insns.split(",")]
        insns = [knl.id_to_insn[i] for i in insns]

    # 2. Writers of quantities that depend on the inames and that 1) need -> write_insns
    #
    # To determine these, we need to
    # * Find all written quantities in the instructions from 1)
    # * Find the instructions that write to these quantities
    # * Filter only those that depend on the given inames
    quantities = []
    for insn in insns:
        quantities.extend(insn.read_dependency_names() - inames)

    write_match = lp.match.Or(tuple(lp.match.Writes(q) for q in quantities))
    iname_match = lp.match.And(tuple(lp.match.Iname(i) for i in inames))
    match = lp.match.And((write_match, iname_match))
    write_insns = lp.find_instructions(knl, match)

    # 3. "Outside" instructions, that the instructions from 1) depend on -> dep_insns
    deps = frozenset([]).union(*(insn.depends_on for insn in insns + write_insns))
    deps = deps - frozenset([i.id for i in insns + write_insns])
    dep_insns = lp.find_instructions(knl, lp.match.Or(tuple(lp.match.Id(i) for i in deps)))

    # 4. All other instructions (needed to correctly reconstruct) -> other_insns
    used_ids = frozenset([i.id for i in insns + write_insns + dep_insns])
    other_insns = lp.find_instructions(knl, lp.match.Not(lp.match.Or(tuple(lp.match.Id(i) for i in used_ids))))

    #
    # Assert some assumptions on the instructions
    #

    # All instructions within the given inames are either to be vectorized
    # or write a dependent quantity
    for insn in other_insns + dep_insns:
        len(insn.within_inames.intersection(inames) == 0)

    # An instruction occurs in but one of these groups:
    assert len(set(insns + write_insns + dep_insns + other_insns)) == len(insns + write_insns + dep_insns + other_insns)

    # All the target and write instructions are Assignments
    for insn in insns + write_insns:
        assert isinstance(insn, lp.Assignment)

    # Analyse the inames of the given instructions and identify inames
    # that they all have in common. Those inames will also be iname dependencies
    # of inserted instruction.
    common_inames = frozenset([]).union(*(insn.within_inames for insn in insns)) - inames

    # Have a list gathering all *new* instructions
    new_insns = []
    temporaries = knl.temporary_variables

    # Insert a flat consecutive counter 'total_index'
    temporaries['total_index'] = lp.TemporaryVariable('total_index',  # name
                                                      dtype=np.int32,
                                                     )
    new_insns.append(lp.Assignment(prim.Variable("total_index"),  # assignee
                                   0,  # expression
                                   within_inames=common_inames,
                                   within_inames_is_final=True,
                                   depends_on=frozenset(i.id for i in dep_insns),
                                   depends_on_is_final=True,
                                   id="assign_total_index",
                                   ))
    new_insns.append(lp.Assignment(prim.Variable("total_index"),  # assignee
                                   prim.Sum((prim.Variable("total_index"), 1)),  # expression
                                   within_inames=common_inames.union(inames),
                                   within_inames_is_final=True,
                                   depends_on=frozenset({"assign_total_index"}),
                                   depends_on_is_final=True,
                                   id="update_total_index",
                                   ))

    # Insert a rotating index, that counts 0 , .. , vecsize - 1
    temporaries['rotate_index'] = lp.TemporaryVariable('rotate_index',  # name
                                                       dtype=np.int32,
                                                       )
    new_insns.append(lp.Assignment(prim.Variable("rotate_index"),  # assignee
                                   0,  # expression
                                   within_inames=common_inames,
                                   within_inames_is_final=True,
                                   depends_on=frozenset(i.id for i in dep_insns),
                                   depends_on_is_final=True,
                                   id="assign_rotate_index",
                                   ))
    new_insns.append(lp.Assignment(prim.Variable("rotate_index"),  # assignee
                                   prim.Remainder(prim.Sum((prim.Variable("rotate_index"), 1)), vec_size),  # expression
                                   within_inames=common_inames.union(inames),
                                   within_inames_is_final=True,
                                   depends_on=frozenset({"assign_rotate_index"}),
                                   depends_on_is_final=True,
                                   id="update_rotate_index",
                                   ))

    # Pre-evaluate all the needed quantities
    for quantity in quantities:
        name = quantity + '_buffered'
        temporaries[name] = lp.TemporaryVariable(name,
                                                 dtype=np.float64,
                                                 shape=(vec_size,),
                                                 )

    for insn in write_insns:
        # Dummy
        new_insns.append(insn.copy(assignee=prim.Subscript(prim.Variable(name), (prim.Variable('rotate_index'),)),
                                   )
                         )

    # Insert the 'continue' statement
    new_insns.append(lp.CInstruction((),  # iname exprs that the code needs access to
                                     "continue;",  # the code
                                     predicates=frozenset({"rotate_index == 0", "blubb"}),
                                     depends_on=frozenset({"update_rotate_index", "update_total_index"}).union(frozenset([i.id for i in write_insns])),
                                     depends_on_is_final=True,
                                     within_inames=common_inames.union(inames),
                                     within_inames_is_final=True,
                                     id="continue_stmt",
                                     ))

    #
    # TODO! Replace the precomputed quantities in the target instructions
    #

    for insn in insns:
        new_insns.append(insn.copy(expression=insn.expression))

    # Return a kernel
    return knl.copy(instructions=dep_insns + other_insns + new_insns,
                    temporary_variables=temporaries,
                    )
